﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("WorstQualityAnalyzer", "An operator that collects the worst quality found.")]
  [StorableClass]
  public class WorstQualityAnalyzer : SingleSuccessorOperator, IStatefulItem {
    private const string GenerationsParameterName = "Generations";
    private const string ResultsParameterName = "Results";

    public IValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> ParentsQualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["ParentsQuality"]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public ILookupParameter<DoubleValue> WorstKnownQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["WorstKnownQuality"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    [Storable]
    private double upperBound = double.MinValue;

    [StorableConstructor]
    private WorstQualityAnalyzer(bool deserializing) : base(deserializing) { }
    private WorstQualityAnalyzer(WorstQualityAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      upperBound = (double)original.upperBound;
    }

    public WorstQualityAnalyzer()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ParentsQuality", "The quality of the parent solutions."));
      ParentsQualityParameter.ActualName = "TSPTourLength";
      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, false otherwise"));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));
      Parameters.Add(new LookupParameter<DoubleValue>("WorstKnownQuality", "The quality of the worst known solution of this problem."));
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new WorstQualityAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var qualityParent1 = ParentsQualityParameter.ActualValue.First().Value;
      var qualityParent2 = ParentsQualityParameter.ActualValue.Last().Value;

      if (GenerationsParameter.ActualValue.Value == 0) {
        if (MaximizationParameter.ActualValue.Value) {
          if (upperBound > qualityParent1) {
            upperBound = qualityParent1;
          }
          if (upperBound > qualityParent2) {
            upperBound = qualityParent2;
          }
        } else {
          if (upperBound < qualityParent1) {
            upperBound = qualityParent1;
          }
          if (upperBound < qualityParent2) {
            upperBound = qualityParent2;
          }
        }
      } else {
        if (WorstKnownQualityParameter.ActualValue == null) {
          AddVariableToGlobalScope("WorstKnownQuality", "The quality of the worst known solution of this problem.", new DoubleValue(upperBound));
          ResultsParameter.ActualValue.Add(new Result("WorstQuality", new DoubleValue(upperBound)));
        }
      }

      return base.Apply();
    }

    public override void ClearState() {
      upperBound = double.MinValue;
    }

    public override void InitializeState() { }

    private void AddVariableToGlobalScope(string name, string description, DoubleValue value) {
      var globalScope = ExecutionContext;
      while (globalScope.Parent != null) {
        globalScope = globalScope.Parent;
      }

      globalScope.Scope.Variables.Add(new Variable(name, description, value));
    }
  }
}
